/*===================================================================
  Xwindow Realtime Simulator ver.1 (1994.12.27)
  by kitsune(originated by sasa (originated by  gutchi (originated in WTNB)))
===========================================================================*/

//************************** include *********************************

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>


//************************** define ***********************************


//--------- constant ( must not be changed) ---------

#define pi      3.14159265358979
#define MCOLOR    65535
#define PHI_RED   (pi/3.0)       // RED phase                                  
#define PHI_GREEN (pi)           // GREEN phase                                
#define PHI_BLUE  (-pi/3.0)      // BLUE phase                                 
          
#define SAMPLE    WINMAX         // window number(=w) of sample window in XRS  


//--------- constant ( may be changed) ---------

#define FONT       "6x13"        //text font

#define FXNUM       7      // character number of event & parameter name
#define FXNUMD    "%7.7lf" // ="% FXNUM .* lf" print format of double variable 
#define FXNUMI    "%7d"    // ="% FXNUM d"     print format of int variable
#define FX     (7*FXNUM+1) // width of each function region in control window
#define FY          20     // height of each function region in control window 
#define TITLENUM    16     // character number of graphic window title

#define SWINTITLE  "Sample Window"  //sample window title
#define OWINTITLE  "Control Window" //control window title


//-------- maximam value ( change,as you like ) --------

#define MAX         502    // array size(x direction)                         
#define MAY         502     // array size(y direction) 
#define EVENTMAX     10     // event number(=max of eventnum)
#define PARAMAX      15     // parameter number(=max of paranum)
#define DISPMAX       5     // display number by a window(=max of titlenum[w])
#define WINMAX       10     // graphic window number(=max of winnum)
#define MAPMAX        2     // map array number of graphic window
                            //                      (=max of mapwinum)
#define COLORMAX    146     // color number(=max of colornum[w])


//-------- default value ( can be changed in your program ) -----------

#define WINDOWSIZEi   100       // default window size(x derection)
#define WINDOWSIZEj   100       // default window size(y derection)
#define SWINDOWSIZEi  100       // default sample window size(x derection)
#define SWINDOWSIZEj   50       // default sample window size(y derection)
#define LATTICESIZE     2       // default lattice size of window

#define COLORTONE     'M'       // default color tone  
#define COLORNUMBER   12        // default color number 
#define PHI_0     (-pi*2.0/5.0) // default color OFFSET phase 

#define AMPF          0.8       // default amplitude in defaultdraw function 
#define DPARAM          1       // default PARAMETER CHANGE unit size 


//-------- simple figure draw functions ----------------

#define Line(w,i,j,k,l) XDrawLine(XRSname.d,XRSname.win[w],XRSname.bgc,i,j,k,l)
#define WhiteLine(w,i,j,k,l) \
        XDrawLine(XRSname.d,XRSname.win[w],XRSname.wgc,i,j,k,l)
#define Dash(w,i,j,k,l) XDrawLine(XRSname.d,XRSname.win[w],XRSname.dgc,i,j,k,l)
#define Point(w,i,j)    XDrawPoint(XRSname.d,XRSname.win[w],XRSname.bgc,i,j)
#define WhitePoint(w,i,j)  XDrawPoint(XRSname.d,XRSname.win[w],XRSname.wgc,i,j)
#define Rectangle(w,i,j,x,y) \
        XDrawRectangle(XRSname.d,XRSname.win[w],XRSname.bgc,i,j,x,y)
#define Circle(w,i,j,r)      \
        XDrawArc(XRSname.d,XRSname.win[w],XRSname.bgc,i-r,j-r,2*r,2*r,0,360*64)
#define String(w,i,j,s)      \
        XDrawString(XRSname.d,XRSname.win[w],XRSname.bgc,i,j,s,strlen(s))
               

               
//*************** struct & class & function declare **************************

class     XRS;

struct argi_manip{               //struct for 1 argument(int) manipurator
          void  (*fp)(XRS&,int);
          int   arg;
          argi_manip(void (*f)(XRS&,int),int a){fp=f;arg=a;}
};

struct argd_manip{               //struct for 1 argument(double) manipurator
          void  (*fp)(XRS&,double);
          double x;
          argd_manip(void (*f)(XRS&,double),double y){fp=f;x=y;}
};

struct argfi_manip{              //struct for 2 argument(float,int) manipurator
          void  (*fp)(XRS&,float,int);
          float x;
          int   arg;
          argfi_manip(void (*f)(XRS&,float,int),float y,int a){
                           fp=f;  x=y;  arg=a;}
};

struct argfunci_manip{    //struct for 2 argument(unsigned long (*)(int),int)
          void  (*fp)(XRS&,unsigned long (*)(int), int);
          unsigned long (*func)(int);
          int   arg;
          argfunci_manip(void (*f)(XRS&,unsigned long (*)(int),int),
                   unsigned long (*g)(int),int a){fp=f; func=g; arg=a;}
};
                                                                   
inline double defaultdraw(double);


//************************* class XRS ************************************

class XRS{

      int     i_max[WINMAX+1],j_max[WINMAX+1];
      int     *changedisp[WINMAX],px[WINMAX];
      int     paranum,eventnum;
      int     winevent[WINMAX],titlenum[WINMAX],colornum[WINMAX+1];
      int     id,w,winum,mapwinum,unmapwinum,funcnum,ke,swinopen;
      int     *ts; 

      float   phi0[WINMAX+1];

      double  *t;
      double  *p[PARAMAX],dp[PARAMAX];
   
      void    (*eventfunc[EVENTMAX])(int);
      int     (*wineventfunc[WINMAX])(int);
      double  (*drawfunc[WINMAX+1])(double); 
      unsigned long (*colortone[WINMAX+1])(int);
      
      unsigned long  color[WINMAX+1][COLORMAX]; 

      char    wanya[64];
      char    space[64],pause[64],cont[64],quit[64],start[64];
      char    manip;
      char    tone[WINMAX+1];
      char    eventname[EVENTMAX][FXNUM+1],paraname[PARAMAX][FXNUM+1];
      char    title[WINMAX][DISPMAX][TITLENUM+1];

    public:

      unsigned long  map_old[MAPMAX][MAX][MAY];
      unsigned long  black,white;
	  
      Window   root,owin,win[WINMAX],swin;
      Display  *d;
      GC       gc,bgc,wgc,dgc;
      Colormap cm;
      XEvent   event;
  
     //------------- constractor & destractor ---------------------------
      
      XRS();                                             
      ~XRS(){XCloseDisplay(d);}                         

      
     //------------- definitions for input operator << -------------------
      
      friend XRS& operator<<(XRS& s,XRS& (*f)(XRS &));
                                              //appliciator(for no argument) 
      friend XRS& operator<<(XRS& s,argi_manip m);  // (for 1 argument)
      friend XRS& operator<<(XRS& s,argfi_manip m); // (for 2 argument)
      friend XRS& operator<<(XRS& s,argfunci_manip m);//
      void manipurate(char s){manip=s;}               // set manip  
      void lattice_size(int n){px[w]=n;}              // set lattice size
      void colorstyle(float x,int n,char s)           // set colorstyle 
             { phi0[w]=x; colornum[w]=n; tone[w]=s;}
      void colorfunc( unsigned long (*g)(int))
             { colortone[w]=g; } 
      void swin_open(){swinopen=1;}                   // set sample window open
      void set_unmap(int n){ unmapwinum=n; }          // set unmapwinum
      void set_winum(int n){ if(n+1>winum && n+1<=WINMAX) winum=n+1;}
      void winnumber(int n){if(n<WINMAX+1) w=n; }     // set window number
      XRS& operator<<(double&);                       // 
      XRS& operator<<(int&);                          //  definitions of    
      XRS& operator<<(char name[126]);                //     operator <<
      XRS& operator<<(void (*f)(int));                //
      XRS& operator<<(int  (*f)(int));                //
      XRS& operator<<(double (*f)(double));           //

      
     //-------------- initialize functions --------------------------------

      void  initialize(void)        {           id=0;   window_set();}
      void  initialize(double &time){ t=&time;  id=-1;  window_set();}
      void  initialize(int &step)   { ts=&step; id=1;   window_set();}
      void  window_set();
      void  initialize_array(int);
      void  initialize_color(int);

	  
     //--------------- functions for event --------------------------
    
      int   checkevent();
      void  owin_event();                             // control window
      void  swin_event();                             // sample window
      void  win_event(int);                           // graphic windows


     //--------------- functions for drawing windows -----------------------

      void  owin_draw();                     // control window
      void  swin_draw();                     // sample window 
      void  current_parameter();             // parameter values   
                                             //     in a control window
      void  color_draw(double,int,int,int);  //          (for 2D graphics) 
      void  point_draw(int[],int,int);       // graphic  (for 1D graphics)
      void  point_dot(int,int,int,int);      //  windows (for 1D graphics)  
      void  point_erase(int,int,int);        //          (for 1D graphics)


     //--------------- definition of color & colorstyle ---------------------

      unsigned long color_pixel(float,float,float);
      unsigned long color_pixel(char*);
      unsigned long mono_tone(int,int);       
      unsigned long phase_tone(int,int);

     //--------------- string format function --------------------------------

      void string(char*,char*,int);

     //-------------- function for getting a graphic context ----------------

     GC    graphic_context(unsigned long,unsigned int,int,int,int,char*);
	 
};


//*********************** string format function ***********************

void XRS::string(char* s1,char* s2,int n)
{
    register i,j,k;
    char ss[64];

    i=strlen(s2);
    if( i > n ) i=n;
    k=(int)((n-i)/2);
    for( j=0; j<k; j++ ) ss[j]=' ';
    ss[k]='\0';
    strncat(ss,s2,i);
    for(j=k+i; j<n; j++) ss[j]=' ';
    ss[n]='\0';
    
    strcpy(s1,ss);
}


//********* function for getting a graphic context ********************
   
GC XRS::graphic_context(unsigned long color,
	       unsigned int line_width=0,int line_style=LineSolid,
	       int cap_style=CapButt,int join_style=JoinMiter,char* font=FONT)
{
     GC g;
     Font dfont;
     dfont = XLoadFont(d, font);	 
     g = XCreateGC(d, root, 0, 0);
     XSetLineAttributes(d, g, line_width, line_style, cap_style, join_style);
     XSetForeground(d, g, color);
     XSetBackground(d, g, white);
     XSetFont(d, g, dfont);
     return g;
}


//*********************** constractor **********************************

XRS::XRS(void)
{
      d = XOpenDisplay(NULL);

      root = RootWindow(d, 0);

      black = BlackPixel(d, 0); 
      white = WhitePixel(d, 0);

      gc = graphic_context(black,1);
      bgc = graphic_context(black,0);
      wgc = graphic_context(white,0);
      dgc = graphic_context(black,0,LineOnOffDash);
      
//-------------parameter initialize(default value) ----------------

      winum=1;
      unmapwinum=0;
      paranum=0;
      eventnum=0;
          
      for(int k=0;k<PARAMAX;k++) dp[k]=DPARAM;

      static int i=0;                        //graphic window  
      for(k=0;k<WINMAX;k++){
             changedisp[k]=&i;
             i_max[k]=WINDOWSIZEi;
             j_max[k]=WINDOWSIZEj;
             px[k]=LATTICESIZE;
             winevent[k]=0;
             phi0[k]=PHI_0;
             colornum[k]=COLORNUMBER;
             tone[k]=COLORTONE;
             drawfunc[k]=defaultdraw;
      } 

      swinopen=0;                            //sample window
      i_max[SAMPLE]=SWINDOWSIZEi;
	  j_max[SAMPLE]=SWINDOWSIZEj;
      phi0[SAMPLE]=PHI_0;
      colornum[SAMPLE]=COLORNUMBER;
      tone[SAMPLE]=COLORTONE;
      drawfunc[SAMPLE]=defaultdraw; 

}


//************************** manipurator ***********************************

XRS& operator<<(XRS& s,XRS& (*f)(XRS &)){return f(s);}
XRS& operator<<(XRS& s,argi_manip m){m.fp(s,m.arg); return s;}
XRS& operator<<(XRS& s,argfi_manip m){m.fp(s,m.x,m.arg); return s;}
XRS& operator<<(XRS& s,argfunci_manip m){m.fp(s,m.func,m.arg); return s;}

XRS& PARAMETER(XRS& s)
   {s.manipurate('P'); return s;}
XRS& SWINDOW(XRS& s)
   {s.swin_open(); s.manipurate('S'); s.winnumber(SAMPLE); return s;}
XRS& EVENT(XRS& s)
   {s.manipurate('E'); return s;}

void setUNMAP(XRS& s,int n){ s.set_unmap(n); }
argi_manip UNMAP(int n=0){return argi_manip(setUNMAP,n);}
   
void setWINDOW(XRS& s,int n){
       s.manipurate('W'); s.winnumber(n); s.set_winum(n); }
argi_manip WINDOW(int n=0){return argi_manip(setWINDOW,n);}

void setPX(XRS& s,int n){s.lattice_size(n);}
argi_manip PX(int n=LATTICESIZE){return argi_manip(setPX,n);}

void setMONO(XRS& s,float x,int n){s.colorstyle(x,n,'M');}
argfi_manip MONO(float x=PHI_0,int n=COLORNUMBER)
            {return argfi_manip(setMONO,x,n);} 

void setPHASE(XRS& s,float x,int n){s.colorstyle(x,n,'P');}
argfi_manip PHASE(float x=PHI_0,int n=COLORNUMBER)
            {return argfi_manip(setPHASE,x,n);}

void setCOLOR(XRS& s,unsigned long (*g)(int),int n)
            { s.colorstyle(0,n,'C');  s.colorfunc(g);}
argfunci_manip COLOR(unsigned long (*g)(int),int n=COLORNUMBER)
            {return argfunci_manip(setCOLOR,g,n);}
            

//********************** definition of operator << **************************

XRS& XRS::operator<<(int &s)
{
      static int j[WINMAX+1];
      switch(manip){
         case  'W' :{                          //graphic window size input
            if(j[w]==0) i_max[w]=s;
            else if(j[w]==1) j_max[w]=s;
            else if(j[w]==2) changedisp[w]=&s; //pointer of changedisp input 
            j[w]++;
            return *this;
         }
         case  'S' :{                          //sample window size input
            if(j[w]==0)      i_max[w]=s;
		    else if(j[w]==1) j_max[w]=s;
			j[w]++;
            return *this;
     }
         default:{
            return *this;
         }
     }  
}     
            

XRS& XRS::operator<<(double& s)
{
      switch(manip){
         case   'P'  :{                        //pointers of parameter input 
            p[paranum]=&s;
            ++paranum;
            return *this;
         }
         default     :{
            return *this;
     }
      } 
}


XRS& XRS::operator<<(char name[64])
{
      static int i,j;
      switch(manip){
         case   'P'  :{                        //parameter name input
            string(paraname[i],name,FXNUM);
            ++i;
            return *this;
         }
         case   'W'  :{                        //graphic window titl input
            if(w<WINMAX){
               string(title[w][titlenum[w]],name,TITLENUM);
               ++titlenum[w];
            }
            return *this;
         }
         case   'E'  :{                        //event names input
            string(eventname[j],name,FXNUM);
            ++j;
            return *this;
         }
         default     :{
            return *this; 
         }
      }
}


XRS& XRS::operator<<(void (*f)(int))
{
      switch(manip){
         case   'E'  :{                    //pointers of event function input
            eventfunc[eventnum]=f;
            ++eventnum;
            return *this;
         }
         default     :{
            return *this;
         }
      }
}


XRS& XRS::operator<<(int (*f)(int))
{
      switch(manip){
         case   'W'  :{
            wineventfunc[w]=f;
            winevent[w]=1;
            return *this;
         }
         default     :{
            return *this;
         }
      }
}
      

XRS& XRS::operator<<(double (*f)(double))
{
      switch(manip){
         case   'W'  :{
            drawfunc[w]=f;                 //pointers of draw function input
            return *this;
         }
         case    'S' :{
            drawfunc[SAMPLE]=f;
            return *this;
         }
         default     :{
            return *this;
         }
      }
}



//***************** window_setting & initialize ****************************

void XRS::window_set(void)
{
      mapwinum=winum-unmapwinum;
      if(mapwinum<0) mapwinum=0;
      if(mapwinum>=MAPMAX) mapwinum=MAPMAX;
          
//------- graphic window set -------

      for(int k=0;k<winum;k++){
             win[k] = XCreateSimpleWindow(d, root, 300, 300,i_max[k]*px[k]
                               ,j_max[k]*px[k], 2, black, white);
             XStoreName(d, win[k], title[k][*changedisp[k]]);
             XSelectInput(d, win[k] , ButtonPressMask|ExposureMask );
             XMapRaised(d, win[k]);
             initialize_color(k); 
             if(k<mapwinum) initialize_array(k);
      }

//------- sample window set --------

      if(swinopen){
         swin = XCreateSimpleWindow(d, root, 500, 300, i_max[SAMPLE],
                               j_max[SAMPLE], 2, black, white);
         XStoreName(d, swin, SWINTITLE);
         XSelectInput(d, swin, ExposureMask);
         XMapRaised(d,swin);
         initialize_color(SAMPLE);
     }

//------- control window set-----------

	  if(id){
          funcnum=2+1+eventnum+(1+eventnum)%2+2*paranum;
          ke=1+(1+eventnum+(1+eventnum)%2)/2;
	  }else{
		  funcnum=2+eventnum+eventnum%2+2*paranum;
		  ke=1+(eventnum+eventnum%2)/2;
	  }
     
      owin = XCreateSimpleWindow(d, root, 100, 50, FX*(funcnum/2)+1,
                       FY*2+1,2, black, white);
      XStoreName(d, owin, OWINTITLE);
      XSelectInput(d, owin, StructureNotifyMask);
      XMapRaised(d, owin);

//------- window initialization -----

      string(space,""     ,FXNUM);
      string(start,"START",FXNUM);
      string(quit ,"QUIT" ,FXNUM);
      string(pause,"PAUSE",FXNUM);
      string(cont ,"CONT" ,FXNUM);

      while(event.type != MapNotify)
          XWindowEvent(d, owin, StructureNotifyMask, &event);

      XSelectInput(d, owin,ButtonPressMask|ExposureMask);
      if(swinopen) swin_draw();

      owin_draw();
      current_parameter();
      XDrawImageString(d, owin, bgc,    0+5, FY-4, start, FXNUM);

      XFlush(d);
   
      XNextEvent(d, &event);
      while(event.type != ButtonPress)
          XNextEvent(d, &event);
      XDrawImageString(d, owin, bgc,    0+5, FY-4, quit, FXNUM);
}


//********************* initialize map **************************

void  XRS::initialize_array(int k)
{
      register int i,j;

      for(j=0; j<MAY; j++)
          for(i=0; i<MAX; i++)
              map_old[k][i][j] = white;
}


//******************** initialize color ************************

void  XRS::initialize_color(int w)
{
      int i;
      XColor   cc;

      cm = DefaultColormap(d, 0);

      for(i=0; i<colornum[w]; i++){ 
         if     (tone[w]=='M') color[w][i]=mono_tone(i,w);
         else if(tone[w]=='P') color[w][i]=phase_tone(i,w);
         else if(tone[w]=='C') color[w][i]=colortone[w](i);
     }      

      cc.red=0;                          // black
      cc.green=0;
      cc.blue=0;
      XAllocColor(d, cm, &cc);
      color[w][colornum[w]] = cc.pixel;

      cc.red=MCOLOR;                     // white
      cc.green=MCOLOR;
      cc.blue=MCOLOR;
      XAllocColor(d, cm, &cc);
      color[w][colornum[w]+1] = cc.pixel;
}


//************************* event *******************************

inline int XRS::checkevent(void)
{
      register int k,re;

	  re=0;
      if(XCheckWindowEvent(d, owin, ButtonPress|ExposureMask , &event))
		  { owin_event(); re|=1; }
      if(XCheckWindowEvent(d, swin, ExposureMask , &event))
          { swin_event(); re|=2; }
      
      for(k=0;k<winum;k++)
          if(XCheckWindowEvent(d, win[k], ButtonPress|ExposureMask , &event))
             { win_event(k); re|=4; re|=k<<3; }

	  if(id){
         XDrawImageString(d, owin, bgc, FX+5, FY-4,space, FXNUM);
         if(id==-1) sprintf(wanya, FXNUMD,*t);
         else      sprintf(wanya, FXNUMI,*ts);     
         XDrawImageString(d, owin, bgc, FX+5, FY-4, wanya, FXNUM);
	  }
	  return re;
}


//*********************** control window event ***************************** 

void  XRS::owin_event(void)
{
      if( event.type == Expose ){ owin_draw(); return;}

      int fun,xfun,yfun,funid;
          
      xfun=event.xbutton.x/FX;
      yfun=event.xbutton.y/FY;
      fun=2*xfun+yfun;
      if(id) funid=1; else funid=0;
	  
      if (fun == 0){
          /* Quit */
          exit(1);
      
      } else if (fun == 1){
          /* Pause/Cont */
          owin_draw();
          XDrawImageString(d, owin, bgc,   5,   2*FY-4,cont, FXNUM);
          XWindowEvent(d, owin, ButtonPress,  &event);
          XDrawImageString(d, owin, bgc,   5,   2*FY-4,pause, FXNUM);

      } else if (fun == 1+funid){
          /* Time or Step value */
          if(id==-1)     *t=0;
          else if(id==1)  *ts=0;

      } else if (fun>=2+funid && fun<eventnum+2+funid ){
          /* each eventfunction */
          eventfunc[fun-2-funid](event.xbutton.button);  
          current_parameter();

      } else if (fun >= 2*ke && fun < funcnum){
          /* parameter change */
          
          if (yfun == 0){
              /* parameter up */
              if (event.xbutton.button==1)
                  *p[xfun-ke]+=dp[xfun-ke];
              else if (event.xbutton.button==2)
                  *p[xfun-ke]+=dp[xfun-ke]*0.1;
              else if (event.xbutton.button==3)
                  dp[xfun-ke]=dp[xfun-ke]*10.0;
              current_parameter();
 
          }else{
              /* parameter down */
              if (event.xbutton.button==1)
                  *p[xfun-ke]-=dp[xfun-ke];
              else if (event.xbutton.button==2)
                  *p[xfun-ke]-=dp[xfun-ke]*0.1;
              else if (event.xbutton.button==3)
                  dp[xfun-ke]=dp[xfun-ke]*0.1;
              current_parameter();
 
          }
     
      } else {
      }
}
         

//****************** sample window event ************************

void  XRS::swin_event(void){  swin_draw(); }
          

//****************** graphic window event ************************

void  XRS::win_event(int k)
{
    if(  event.type == Expose ){
        register i,j,ip,jp;
        if( k<mapwinum )
            for(i=0; i<i_max[k]; i++)
                for(j=0; j<j_max[k]; j++){
                    XSetForeground(d, gc, map_old[k][i][j]);
                    for(ip=0; ip<px[k]; ip++)
                       for(jp=0; jp<px[k]; jp++)
                           XDrawPoint(d, win[k], gc, i*px[k]+ip, j*px[k]+jp);
                }
        return;
    }

    if( winevent[k] )
        if( wineventfunc[k](event.xbutton.button) ) return;
    
    switch(event.xbutton.button){
         case 1 :{
             /* display button click location in a graphic window */
             
             printf("Window %d   Click at (i,j)=(  %d , %d ) ",
                    k,event.xbutton.x/px[k],event.xbutton.y/px[k]);

			 if(id){
                 if(id==-1){  printf(" time = ");      printf(FXNUMD,*t); }
                 else     {  printf(" time step = "); printf(FXNUMI,*ts);}
             }
		     printf("\n");
             break;
         }
         case 2 :{
             /* display change */

             if (*changedisp[k]==(titlenum[k]-1))
                *changedisp[k]=0; 
             else
                ++(*changedisp[k]); 
			 XStoreName(d, win[k], title[k][*changedisp[k]]);
             break;
         }
         default :{
         }
    } 
}



//******************** control window draw **************************

void  XRS::owin_draw(void)
{
      register int i,j,k,rec,funid;

      XDrawLine(d, owin, bgc, 0, FY, FX*(funcnum+1)/2+1, FY);
      for (i=1;i<funcnum/2;i++){
           XDrawLine(d, owin, bgc, FX*i, 0, FX*i, FY*2);
      }

      XDrawImageString(d, owin, bgc,    0+5,   FY-4,quit, FXNUM);
      XDrawImageString(d, owin, bgc,    0+5, 2*FY-4,pause, FXNUM);

	  if(id) funid=1; else funid=0;
      for (i=0; i<eventnum; ++i){ 
          j=(i+funid)%2+1; k=1+((i+funid)-(i+funid)%2)/2;
          XDrawImageString(d, owin, bgc, k*FX+5, j*FY-4, eventname[i], FXNUM);
      }

      for(i = 0; i<paranum; ++i){
          k=i+ke;
          XDrawImageString(d, owin, bgc, k*FX+5,   FY-4,  paraname[i], FXNUM);
          rec=sprintf(wanya, FXNUMD,*p[i]);
          XDrawImageString(d, owin, bgc, k*FX+5, 2*FY-4,  wanya, FXNUM);
      }
}


//********************** sample window draw ***************************

void  XRS::swin_draw(void)
{
      register i,j,digA;
      unsigned long  map;
      double sx,mono,sashi,m,s;
      
      mono=(double)i_max[SAMPLE]/2;
      sashi=mono/5;
      
      for( j=1; j<j_max[SAMPLE]; j++){
          for( i=1; i<i_max[SAMPLE]; i++){
 
              m=(double)i-floor((double)i/mono)*mono;
              s=(double)i-floor((double)i/sashi)*sashi;

              sx=(double)i/(double)i_max[SAMPLE];
              digA=(int)(drawfunc[SAMPLE](sx)*colornum[SAMPLE]);
                                
              if ( (int)m==0 && j < 10 )                      /*memori*/
                  map = color[SAMPLE][colornum[SAMPLE]];
              else if ( (int)s==0 && j < 5 )
                  map = color[SAMPLE][colornum[SAMPLE]];
              else                                            /* futsu */
                  map = color[SAMPLE][digA];

              XSetForeground(d, gc, map);
              XDrawPoint(d, swin, gc, i, j);
          }
      }

      XSetForeground(d, gc, white);

}


//************* current parameter draw in a control window ******************

void  XRS::current_parameter(void)
{
    for(int i=0; i<paranum; ++i){
        int k=i+ke; 
        XDrawImageString(d, owin, bgc, k*FX+5, 2*FY-4,space, FXNUM);
        sprintf(wanya, FXNUMD,*p[i]);
        XDrawImageString(d, owin, bgc, k*FX+5, 2*FY-4,  wanya, FXNUM);
    }
    return;
}


//************* color draw in graphic windows ********************************

inline void  XRS::color_draw(double W,int i,int j,int w=0)
{
      register ip,jp,digA;
      unsigned long  map;
      
      digA=(int)(drawfunc[w](W)*colornum[w]);
      if(digA > colornum[w]) digA=colornum[w];
      map = color[w][digA];

      if(w<mapwinum){
          if(map == map_old[w][i][j]) return;
          map_old[w][i][j] = map;
      }

      XSetForeground(d, gc, map);
      for(ip=0; ip<px[w]; ip++)
          for(jp=0; jp<px[w]; jp++)
              XDrawPoint(d, win[w], gc, i*px[w]+ip, j*px[w]+jp);
      
}


//************* point draw (for 1D graphic) **********************************

inline void  XRS::point_draw(int j[],int w=0,int digA=-1)
{
      register i,plot,plotold,imax,jmax;
      unsigned long  map;
      static int plot_old[WINMAX][MAX];

      if( digA < 0 || digA > colornum[w] ) digA=colornum[w];
      map = color[w][digA];

      jmax=j_max[w];
      imax=i_max[w];
      for( i=0; i<imax; i++ ){
          plot=jmax-j[i];
          plotold=plot_old[w][i];
          if(w<mapwinum){
              if(plot==plotold && map == map_old[w][i][plot]) continue;
              map_old[w][i][plotold]=white;
              map_old[w][i][plot]=map;
          }else{
              if( plot==plotold ) continue;
          }
          XDrawPoint(d, win[w], wgc, i, plotold);
          XSetForeground(d, gc, map);
          XDrawPoint(d, win[w], gc, i, plot);
          plot_old[w][i]=plot;
      }
  }


//************* point_dot (for 1D graphic) ************************************

inline void  XRS::point_dot(int i,int j,int w=0,int digA=-1)
{
      register plot;
      unsigned long  map;

      if( digA < 0 || digA > colornum[w] ) digA=colornum[w];
      map = color[w][digA];
      plot=j_max[w]-j;

      if(w<mapwinum){
         if( map == map_old[w][i][plot]) return;
         map_old[w][i][plot]=map;
      }
      XSetForeground(d, gc, map);
      XDrawPoint(d, win[w], gc, i, plot);
}


//************* point_erase (for 1D graphic) **********************************

inline void  XRS::point_erase(int i,int j,int w=0)
{
      register plot;

      plot=j_max[w]-j;
      
      if(w<mapwinum)  map_old[w][i][plot]=white;
      XDrawPoint(d, win[w], wgc, i, plot);
}


//************* color pixel ****************************************

unsigned long XRS::color_pixel(float r,float phi,float brightness=1.0)
{
      float odias, dc, max;
      XColor   cc;

      max=brightness*MCOLOR;
      if( max < 0      ) max=0;
      if( max > MCOLOR ) max=MCOLOR;
      if(   r < 0      ) r=0;
      if(   r > 1      ) r=1;
      
      odias = 1.0-r;
      
      dc=cos(phi-PHI_RED);
      if(dc > 0.5)
         cc.red = (int)max;
      else if (dc < -0.5)
         cc.red = (int)(max*odias);
      else
         cc.red = (int)(max*(odias+(1.0-odias)*(dc+0.5)));

      dc=cos(phi-PHI_GREEN);
      if(dc > 0.5)
         cc.green = (int)max;
      else if (dc < -0.5)
         cc.green = (int)(max*odias);
      else
         cc.green = (int)(max*(odias+(1.0-odias)*(dc+0.5)));

      dc=cos(phi-PHI_BLUE);
      if(dc > 0.5)
         cc.blue = (int)max;
      else if (dc < -0.5)
         cc.blue = (int)(max*odias);
      else
         cc.blue = (int)(max*(odias+(1.0-odias)*(dc+0.5)));

      XAllocColor(d, cm, &cc);
      return cc.pixel;
}


unsigned long XRS::color_pixel(char* name)
{
      XColor   cc;

      if(!XParseColor(d,cm,name,&cc)){
          printf("Color request fails. No color has a such name\n");
          exit(1);
      }

      XAllocColor(d, cm, &cc);
      return cc.pixel;
}

	  
//*********** mono/phase tone *************************************

unsigned long XRS::mono_tone(int i,int w){
      return color_pixel((float)(i)/colornum[w],phi0[w]);
}
      
unsigned long XRS::phase_tone(int i,int w){
      return color_pixel(1,2*pi*(i+0.5)/colornum[w]+phi0[w]);
}


//********** default draw function ********************************

inline double defaultdraw(double x){ return(x*x*AMPF);}


#undef MCOLOR
#undef PHI_RED
#undef PHI_GREEN
#undef PHI_BLUE
#undef FX
#undef FONT
#undef FXNUM
#undef FXNUMD
#undef FXNUMI
#undef FY
#undef TITLENUM
#undef SWINTITLE
#undef OWINTITLE
#undef EVENTMAX
#undef PARAMAX
#undef DISPMAX
#undef WINMAX
#undef MAPMAX
#undef COLORMAX
#undef WINDOWSIZEi
#undef WINDOWSIZEj
#undef SWINDOWSIZEi
#undef SWINDOWSIZEj
#undef LATTICESIZE
#undef COLORTONE
#undef COLORNUMBER
#undef PHI_0
#undef AMPF
#undef DPARAM










