/*==========================================================================
   Xwindow Realtime Simulator ver.1 Sample (2D simulation)
   by kitsune(originated by sasa (originated by  gutchi (originated in WTNB)))
==========================================================================*/

/*------------------------ comment ----------------------------------

   lines for XRS in this program are marked X at the top.   

   improvements of XRSver1.h
    ! Ver.1 input window size and changedisp by << , not by X.initialize(,,,).
    * More than one graphic windows can be opened.
    * Display changes by putting on a mouse botton in the graphic window. 
    * User events and draw function can be made.
    * Two colorstyle (monotone,phasetone) can be used in each window.
         ( It is easy to add user colorstyle in XRSver1.h.)
    * Almost parameters can be changed. (They have <default values>.)    
    - But, complex type of C++ can not be used.       

 * define
      #define pi 3.14159265358979
      #define MAX        202        maximam array size
      #define MAY        202
      #define EVENTMAX    10        maximam event number
      #define PARAMAX     10        maximam parameter number
      #define WINMAX      10        maximam window number
      #define MAPMAX       2        maximam number of window using map array
      #define COLORMAX   12*12+2    maximam color number
      #define <default value>

      are wrriten in a include file "XRS.h".
      you may use or change, as you like.

 * classes defined in XRS.h
     class XRS;       

 * main menbers of class XRS
     XRS()      : constractor of class XRS.

     operator<< : operator for window parameters input
        manipurator 
           EVENT        <<"event name"(char s[8])
                        <<event function(void (*f)(int n))
                           n is a kind of a mouse button putting on.
           PARAMETER    <<"parameter name"(char s[8])
                        <<parameter pointer(double)
           UNMAP(int n)  : last n windows don't use map array. <default 0>            
           WINDOW(int n) : n'th graphic window opens.
                        <<"display name"(char s[16])
                        <<window x size in graphic window n (int) <WINDOWSIZEi>
                        <<       y                          (int) <WINDOWSIZEj>
                        <<pointer of changedisp(int)               <0> 
                        <<pointer of draw function(double (*f)(double)) 
                             <double defaultdraw(double W)=W*W*AMPF>
           SWINDOW       : sample window opens.
                        <<sample window size(int)                  <SWINS>
           ---- Next manipurators follow on WINDOW(n) or SWINDOW. ------- 
           PX(int n) : lattice size of the window <LATTICESIZE>
           MONO(float phi,int c) : set colorstyle of the window monotone 
                           phi is phase of color, c is color number
           PHASE(float phi,int c): set colorstyle of the window phasetone
                           phi is initial phase of color, c is color number
                         <COLORTONE,PHI_0,COLORNUMBER>

     void initialize(double& or int&);
               : initialize windows and set pointer of t,
               :   t is acceptable to be either type , double or int.  
     void checkevent();                      
               : check event and display t.
     void color_draw(double W,int i,int j,int w); (for 2D simulation)
               : draw color at point(i*PX,j*PX) in a graphic window w (<0>).
               : color = (int)(color number * draw function f(W))
               : 0<=W<1 , W>=1 --> color is black


     The following functions return displayID,windowID,GC etc using in XRS.  
       ( These are functions for using Xwindow commands in your program.)
        
      Display*         display();                      displayID.

      Window       root_window();                 root window ID.
      Window    control_window();              control window ID.
      Window     sample_window();               sample window ID.
      Window            window(int w);    w'th graphic window ID.

      GC            colorsolid();         graphic context for 2D simulation   
      GC            blacksolid();         graphic context of black solid line
      GC            whitesolid();         graphic context of white solid line
      GC             blackdash();         graphic context of black dashed line
      Font                font();         text font 
      XEvent            xevent();         event union using in checkevent()

 * Functions drawing simple figures are definded by "#define" in XRSver1.h.
   w is window number.
      Line(w,i,j,k,l)      : a black solidline between point(i,j) and (k,l) 
      WhiteLine(w,i,j,k,l) : a white solidline
      Dash(w,i,j,k,l)      : a black dashed line
      Point(w,i,j)         : a black point at (i,j) 
      Rectangle(w,i,j,x,y) : a rectangle with wide x and height y at (i,j) 
      Circle(w,i,j,r)      : a circle with center (i,j) and radius r
      String(w,i,j,s)      : a text string s at (i,j)
   These aren't best at the poit of calculation speed.
 

  ------------------------- XRSver1sample.C -----------------------
  
  XRSver1sample.C simulates Kobayashi Model of crystal growth.
  (This program is written in C except some lines(C++) for XRS.
   If you don't use graphics, you set CC 1,
   then this program can be compiled as C.)   
    
     tau*e*e*dP/dt = e*e*Laplacian(P)+P*(1-P)*(P-0.5+m(T)+disturb*RANSU)
     dT/dt = Laplacian(T)+K*dP/dt
     m(T) = alpha*Arctan(beta*(1-T))/pi 

  how to run
     XRSver1sample pqr parameterfile1 [parameterfile2 parameterfile3 ...]
         p=B -> Background
         q=y -> snapshot file save
         r=a -> load snapshot file at first 
   
---------------------------------------------------------------------------*/


/*********** include **************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#if !CC
/*X*/  #include "/home/home2/kitsune/include/XRSver1.h"
#endif


/*********** define ***************/

#if CC
#define pi      3.14159265358979
#define MAX         202         /* array size(x direction) */                  
#define MAY         202         /* array size(y direction) */
#endif

#define M(x)        (alpha*atan(beta*(1-x))/pi)
#define RAN()       ((double)rand()/2147483647)
#define RANSU       (RAN()-0.5)
#define integer(x)  ((int)(floor(x)+floor(2*((x)-floor(x)))))


/*********** global variable ************/

double                  t;
double                  P[MAX][MAY];
double                  T[MAX][MAY];

double                  LX,LY;
double                  DT,DX;
double                  time_interval;
double                  save_time;
double                  quit_time;

double                  alpha;
double                  beta;
double                  tau;
double                  K;
double                  e;
double                  disturb;
double                  r_start;

char                    name[26]; /* snapshot_save filename */
 
int                     i_max,j_max;
int                     changedisp;   

FILE                    *file;

/********** function & class prototype *****************/

void                    read_parameter(char *filename);
void                    initialize_function(void);
void                    snapshot_save(char*);
void                    snapshot_load(char*);
void                    n_snapshot_save(int n);
void                    boundary_condition(void);
void                    win_draw(void);

#if !CC
void                    save_event(int);
void                    load_event(int);
void                    initial_event(int);

/*X*/  XRS              X;
#endif   


/********** main ***********/

main(int argc,char *argv[])
{

#if !CC
/*----------------- help -------------------*/
   if( argc!=3 ){
      printf("  how to run \n");
      printf("\n");
      printf("   XRSver1sample pqr parameterfile1 [parameterfile2 parameterfile3 ...] \n");
      printf("   p=='B' -> not open windows for Background job.                       \n");
      printf("   q=='y' -> save snapshot files.                                       \n");
      printf("   r=='a' -> load a snapshot file at first.                             \n");
      printf("     for example, XRSver1sample Fnn XRSver1sample.para                  \n");
      printf("\n");
      printf("   Please read simple comment of XRSver1.h written in 'XRSver1sample.C'.\n");
      exit(0);
   }
#endif


/*----- main routine variable -------*/
 
     int   step,disp_step,save_step,quit_step,paranumber;
     int   Fore,Save,Append;

     Fore  = (* argv[1]   !='B');
     Save  = (*(argv[1]+1)=='y');
     Append= (*(argv[1]+2)=='a');


/*----- parameter input ------*/
    
     for(paranumber=2;paranumber<argc;paranumber++){

        read_parameter(argv[paranumber]);

        i_max = (int)(LX/DX) + 1;
        j_max = (int)(LY/DX) + 1;
        if(Fore) time_interval=DT; else time_interval=save_time; 
        disp_step=integer(time_interval/DT);
        save_step=integer(save_time/DT);
        quit_step=integer(quit_time/DT);


/*----- initialize ----------*/

#if !CC
/*X*/   X<<EVENT    <<   "initial" << "SAVE"<< "LOAD"
/*X*/               <<initial_event<<save_event<<load_event
/*X*/    <<PARAMETER<< alpha << beta << K << e
/*X*/               <<"alpha"<<"beta"<<"K"<<"e"
/*X*/    <<WINDOW(0)<<"P"<<"T"
/*X*/               <<i_max<<j_max<<changedisp
/*X*/    <<SWINDOW  <<50<<50;
                        
     
/*X*/   if(Fore){ X.initialize(t);  win_draw();   }
#endif

        initialize_function();

#if !CC
        if(Append) load_event(0);
#endif


/*----- main routine --------*/

        while(1){

            int ts;

#if !CC      
/*X*/       if(Fore) X.checkevent();
#endif

            double   T2[MAX][MAY],P2[MAX][MAY],P3[MAX][MAY];
            double   nl,dP,dT,tee,txx,txxp;
            tee=DT/(e*e*tau);
            txx=DT/(DX*DX);
            txxp=txx/tau;

 
            for(ts=0; ts<disp_step; ts++){

                register i,j;
 
/*----- P local map -----*/
                for(i=1;i<i_max;i++)
                    for(j=1;j<j_max;j++){
                        nl=P[i][j]*(1-P[i][j])
                          *(P[i][j]-0.5+M(T[i][j])+disturb*RANSU);
                        P2[i][j]=P[i][j]+tee*nl;
                    }

/*----- P & T diffusion term -----*/

                for(i=1;i<i_max;i++)
                    for(j=1;j<j_max;j++){
                        dP=P[i-1][j]+P[i+1][j]+P[i][j-1]+P[i][j+1]
                            -4.0*P[i][j];
                        P3[i][j]=P2[i][j]+txxp*dP;

                        dT=T[i-1][j]+T[i+1][j]+T[i][j-1]+T[i][j+1]
                            -4.0*T[i][j];

                        T2[i][j]=T[i][j]+txx*dT;
                    }

/*----- T local map -----*/
                for(i=1;i<i_max;i++)
                    for(j=1;j<j_max;j++){
                        T[i][j]=T2[i][j]+K*(P3[i][j]-P[i][j]);
                        if( T[i][j] < 0 ) T[i][j]=0.0;
                  
                        P[i][j]=P3[i][j];
                    }

               boundary_condition();
            }

            t = t + time_interval;

            if(Save){
                step=integer(t/DT);
                if((step%save_step)==0){
                    n_snapshot_save((int)(step/save_step));
                    if(Fore) printf("snapshot_save t= %lf \n",t);
                }
                if(step >= quit_step) break;
            }

#if !CC
            if(Fore) win_draw();
#endif
         }
    }
}


/************ boundary condition ************/

void boundary_condition(void)
{
    register int i,j;

/* free + fix */

    for(j=1;j<j_max;j++){
        T[0][j]=T[1][j];
        T[i_max][j]=0;
        P[0][j]=P[1][j];
        P[i_max][j]=0;

    }
    for(i=1;i<i_max;i++){
        T[i][0]=T[i][1];
        T[i][j_max]=0;
        P[i][0]=P[i][1];
        P[i][j_max]=0;
    }
}


/*************** initialize function ************/

void initialize_function(void)
{
    register int i,j;
    double xi;

    for(j=1; j<j_max; j++)
        for(i=1; i<i_max; i++){
            P[i][j] = 0;
            T[i][j] = 0;
        }

    for(j=1; j<j_max; j++)
        for(i=1; i<i_max; i++)
            if(((i-1)*(i-1)+(j-1)*(j-1))<(r_start/DX*r_start/DX)){
                T[i][j] = 1.0;
                P[i][j] = 1.0;
            }

    boundary_condition();

    t=0;
}


#if !CC
/************** window draw **************/

void win_draw(void)
{
    register int i,j;

    if (changedisp==0){

        for(j=1; j<j_max; j++)
            for(i=1; i<i_max; i++)
/*X*/           X.color_draw(P[i][j],i,j);
            
    }else if (changedisp==1){

        for(j=1; j<j_max; j++)
             for(i=1; i<i_max; i++)
/*X*/           X.color_draw(T[i][j],i,j);

    }else{

    }

    return;
}
#endif


/************ number n snapshot_save ************/

void n_snapshot_save(int n)
{
    char filename[126];
    sprintf(filename,"%s%d",name,n);
    snapshot_save(filename);
    return;
}


/*********** snapshot_save **************/

void  snapshot_save(char *filename)
{
      register i,j;
      static int use=0;          /* stop at too many saves */
      ++use;
      if( use > 100 ) exit(1);

      file=fopen(filename,"w");
      
      fprintf(file,"\"t= %lf \"\n",t);
      for(j=1; j<j_max; j++)
          for(i=1; i<i_max; i++)
              fprintf(file,"%d %d %lf %lf\n",i,j,P[i][j],T[i][j]);
      
      fclose(file);
}


/*********** snapshot_load *************/

void  snapshot_load(char *filename)
{
      register i,j;
      int k;
      
      if( NULL==(file=fopen(filename,"r")) ){
          printf(" can't open %s\n",filename);
          exit(1);
      }

      fscanf(file,"%*s%lf%*s",&t);
      for(j=1; j<j_max; j++)
          for(i=1; i<i_max; i++)
              fscanf(file,"%d %d %lf %lf",&k,&k,&(P[i][j]),&(T[i][j]));
      
      fclose(file);
}


/************ read_parameter **************/

void read_parameter(char *filename)
{

    if( NULL==( file=fopen(filename,"r") ) ){
       printf(" can't open %s\n",filename);
       exit(1);
    }

    fscanf(file,"%*s");
    fscanf(file,"%*s%lf",     &LX   );
    fscanf(file,"%*s%lf",     &LY   );
    fscanf(file,"%*s%lf",     &DT   );
    fscanf(file,"%*s%lf",     &DX   );
    fscanf(file,"%*s%lf",   &alpha  );
    fscanf(file,"%*s%lf",   &beta   );
    fscanf(file,"%*s%lf",    &tau   );
    fscanf(file,"%*s%lf",     &K    );
    fscanf(file,"%*s%lf",     &e    );
    fscanf(file,"%*s%lf",  &disturb );
    fscanf(file,"%*s%lf", &save_time);
    fscanf(file,"%*s%lf", &quit_time);
    fscanf(file,"%*s%s",    name    );
    fscanf(file,"%*s%lf",  &r_start );
    fscanf(file,"%*s%d", &changedisp);

    fclose(file); 
}


#if !CC
/*********** event definition ****************/

void save_event(int n)
{
      /* Save */
    char filename[126],buffer[126];
    printf("Enter save filename  ");
    gets(buffer);
    sscanf(buffer,"%s",filename);
    snapshot_save(filename);
    printf("snaphot.data save >> %s \n",filename);
}


void load_event(int n)
{
      /* Load */
    char filename[126],buffer[126];
    printf("Enter load filename  ");
    gets(buffer);
    sscanf(buffer,"%s",filename);
    snapshot_load(filename);
    printf("%s >> snapshot_data load\n",filename);
}

   
void initial_event(int n)
{
     /* initial */
     initialize_function();
}
#endif
          


    


